// Copyright (C) 2008
// by all contributors..

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.


#ifndef SCNSL__RTL__RTLTASK_IF_T_I_HH
#define SCNSL__RTL__RTLTASK_IF_T_I_HH



/// @file
/// Basic rtl task interface.

#include "RtlTask_if_t.hh"
#include "RtlTaskProxy_t.hh"


namespace Scnsl { namespace Rtl {


    // ////////////////////////////////////////////////////////////////
    // Destructor.
    // ////////////////////////////////////////////////////////////////

    template< const int PACKET_SIZE >
    RtlTask_if_t< PACKET_SIZE >::~RtlTask_if_t()
    {
        using namespace sc_core;

        // Calling destructors for all ports.
        // NOTE: reverse order not truly required here,
        // while it is a must in the STL.
        for ( register unsigned int i = _PROXIES_NUMBER; i != 0; )
        {
            --i;

            inputPacket[ i ].~sc_in< rtl_packet_t >();
            outputPacket[ i ].~sc_out< rtl_packet_t >();

            _inputPacketSignal[ i ].~sc_signal< rtl_packet_t >();
            _outputPacketSignal[ i ].~sc_signal< rtl_packet_t >();

			_carrierSignal[ i ].~sc_signal< carrier_t >();
			_inputPacketSizeSignal[ i ].~sc_signal< size_t >();
			_inputPacketLabelSignal[ i ].~sc_signal< label_t >();
			_newInputPacketSignal[ i ].~sc_signal< counter_t >();
			_packetSendCompletedSignal[ i ].~sc_signal< counter_t >();
			_outputPacketSizeSignal[ i ].~sc_signal< size_t >();
			_outputPacketLabelSignal[ i ].~sc_signal< label_t >();
			_newOutputPacketSignal[ i ].~sc_signal< counter_t >();

        }

        // Deallocating the memory.
        ::operator delete[] ( inputPacket );
        ::operator delete[] ( outputPacket );

        ::operator delete[] ( _inputPacketSignal );
        ::operator delete[] ( _outputPacketSignal );

        ::operator delete[] ( _carrierSignal );
        ::operator delete[] ( _inputPacketSizeSignal );
        ::operator delete[] ( _inputPacketLabelSignal );
        ::operator delete[] ( _newInputPacketSignal );
        ::operator delete[] ( _packetSendCompletedSignal );
        ::operator delete[] ( _outputPacketSizeSignal );
        ::operator delete[] ( _outputPacketLabelSignal );
        ::operator delete[] ( _newOutputPacketSignal );

    }


    // ////////////////////////////////////////////////////////////////
    // Implemented support methods for taskporxies.
    // ////////////////////////////////////////////////////////////////

    template< const int PACKET_SIZE >
    typename RtlTask_if_t< PACKET_SIZE >::delete_ports_t
    RtlTask_if_t< PACKET_SIZE >::createAndBoundPorts(
        sc_core::sc_port_base * & inPacket,
        sc_core::sc_port_base * & outPacket,
		RtlTaskProxy_t * tp,
        std::string bindIdentifier )
        
    {
        SCNSL_TRACE_DBG( 5, "<> createAndBoundPorts()." );

        if ( _bounded_proxies >= _PROXIES_NUMBER )
        {
            throw std::logic_error( "Bounding too much proxies." );
        }

        // Creating proxy ports:
		sc_core::sc_out< rtl_packet_t > * in = new sc_core::sc_out< rtl_packet_t >( "inputPacket" );
		sc_core::sc_in< rtl_packet_t > * out = new sc_core::sc_in< rtl_packet_t >( "outputPacket" );

        inPacket = in;
        outPacket = out;

        Scnsl::Core::IdentifierNumber_t newId = Scnsl::Core::IdentifierNumber_t(_bounded_proxies);
        std::string newIdKey;
        std::stringstream ss;
        ss << newId;
        ss >> newIdKey;

        if (bindIdentifier != "") newIdKey = bindIdentifier;

        if(_bindIdMap.find(newIdKey) != _bindIdMap.end())
        {
            throw std::logic_error( "This bind Identifier is already used. Please check bsb bind Identifier attribute." );
        }

	
        // Bounding proxy ports:
        (*in)( _inputPacketSignal[ newId ] );
        (*out)( _outputPacketSignal[ newId ] );

        // Bounding task ports:
        this->inputPacket[ newId ]( _inputPacketSignal[ newId ] );
        this->outputPacket[ newId ]( _outputPacketSignal[ newId ] );

		carrier[ newId ]( _carrierSignal[ newId ] );
		tp->carrier( _carrierSignal[ newId ] );
		inputPacketSize[ newId ]( _inputPacketSizeSignal[ newId ] );
		tp->inputPacketSize( _inputPacketSizeSignal[ newId ] );
		inputPacketLabel[ newId ]( _inputPacketLabelSignal[ newId ] );
		tp->inputPacketLabel( _inputPacketLabelSignal[ newId ] );
		newInputPacket[ newId ]( _newInputPacketSignal[ newId ] );
		tp->newInputPacket( _newInputPacketSignal[ newId ] );

		packetSendCompleted[ newId ]( _packetSendCompletedSignal[ newId ] );
		tp->packetSendCompleted( _packetSendCompletedSignal[ newId ] );
		outputPacketSize[ newId ]( _outputPacketSizeSignal[ newId ] );
		tp->outputPacketSize( _outputPacketSizeSignal[ newId ] );
		outputPacketLabel[ newId ]( _outputPacketLabelSignal[ newId ] );
		tp->outputPacketLabel( _outputPacketLabelSignal[ newId ] );
		newOutputPacket[ newId ]( _newOutputPacketSignal[ newId ] );
		tp->newOutputPacket( _newOutputPacketSignal[ newId ] );


        // Updating bounded proxies:
        ++ _bounded_proxies;
        _bindIdMap[newIdKey] = newId;

        return & delete_ports;
    }

    template< const int PACKET_SIZE >
    void RtlTask_if_t< PACKET_SIZE >::createInternalPacket(
        Packet_t & p,
        sc_core::sc_port_base * outPacket,
        const counter_t outPacketSize )
    {
        SCNSL_TRACE_DBG( 5, "<> createInternalPacket()." );

        // This dynamic cast should never fail.
		// sc_core::sc_in< rtl_packet_t > * pp =
		// dynamic_cast< sc_core::sc_in< rtl_packet_t > * >( outputPacket );
        // Thus, it is replaced with a static cast:
        sc_core::sc_in< rtl_packet_t > * pp =
        static_cast< sc_core::sc_in< rtl_packet_t > * >( outPacket );

        rtl_packet_to_buffer( pp->read(), p, outPacketSize );
    }

    template< const int PACKET_SIZE >
    void RtlTask_if_t< PACKET_SIZE >::writeInternalPacket(
        Packet_t & p,
        sc_core::sc_port_base * inPacket )
    {
        SCNSL_TRACE_DBG( 5, "<> writeInternalPacket()." );

        // This dynamic cast should never fail.
		// sc_core::sc_in< rtl_packet_t > * pp =
		// dynamic_cast< sc_core::sc_in< rtl_packet_t > * >( outputPacket );
        // Thus, it is replaced with a static cast:
        sc_core::sc_out< rtl_packet_t > * pp =
        static_cast< sc_core::sc_out< rtl_packet_t > * >( inPacket );

        rtl_packet_t rtlp;

        byte_t * b;
        p.getPayload( b );
        buffer_to_rtl_packet( b, rtlp, p.getSizeInBytes() );

        pp->write( rtlp );
    }


    // ////////////////////////////////////////////////////////////////
    // General support methods.
    // ////////////////////////////////////////////////////////////////


    template< const int PACKET_SIZE >
    void RtlTask_if_t< PACKET_SIZE >::rtl_packet_to_buffer( const rtl_packet_t & p,
                                                            Packet_t & pck,
                                                            const size_t s )
    {
        const size_t ss = (s%8) + (s/8);
        byte_t * buffer = new byte_t[ ss ];
        for ( register unsigned int i = 0; i < ss; ++i )
        {
            buffer[ i ] = byte_t(p.range( int(PACKET_SIZE - i * 8 -1),
                                   int(PACKET_SIZE - (i + 1) * 8 )).to_int());
        }
        pck.setPayload( buffer, s );
        delete [] buffer;
    }

    template< const int PACKET_SIZE >
    void RtlTask_if_t< PACKET_SIZE >::buffer_to_rtl_packet( const byte_t * buffer,
                                                            rtl_packet_t & p,
                                                            const size_t s )
    {
        for ( register unsigned int i = 0; i < s; ++i )
        {
            p.range( int(PACKET_SIZE - i * 8 -1),
                     int(PACKET_SIZE - (i + 1) * 8) ) =
            static_cast<unsigned int>(buffer[ i ]);
        }
    }


    // ////////////////////////////////////////////////////////////////
    // Protected methods.
    // ////////////////////////////////////////////////////////////////

    template< const int PACKET_SIZE >
    RtlTask_if_t< PACKET_SIZE >::RtlTask_if_t(
        const sc_core::sc_module_name module_name,
        const task_id_t id,
        Node_t * n,
        const size_t proxies )
         :
        // Parent:
        RtlTask_base_if_t( module_name, id, n, proxies ),
        // Ports:
        inputPacket( NULL ),
        outputPacket( NULL ),
        // Internal fields:
        _inputPacketSignal( NULL ),
        _outputPacketSignal( NULL ),
        _carrierSignal( NULL ),
        _inputPacketSizeSignal( NULL ),
        _inputPacketLabelSignal( NULL ),
        _newInputPacketSignal( NULL ),
        _packetSendCompletedSignal( NULL ),
        _outputPacketSizeSignal( NULL ),
        _outputPacketLabelSignal( NULL ),
        _newOutputPacketSignal( NULL )
    {
        // Initializing the ports and signals.
        // Done in this way to permit a good naming.

        // Allocating the memory.
        inputPacket = static_cast< sc_core::sc_in< rtl_packet_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_in< rtl_packet_t >) ));

        outputPacket = static_cast< sc_core::sc_out< rtl_packet_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_out< rtl_packet_t >) ));

        _inputPacketSignal = static_cast< sc_core::sc_signal< rtl_packet_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< rtl_packet_t >) ));

        _outputPacketSignal = static_cast< sc_core::sc_signal< rtl_packet_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< rtl_packet_t >) ));

        _carrierSignal = static_cast< sc_core::sc_signal< carrier_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< carrier_t >) ));

        _inputPacketSizeSignal = static_cast< sc_core::sc_signal< size_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< size_t >) ));

        _inputPacketLabelSignal = static_cast< sc_core::sc_signal< label_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< label_t >) ));

        _newInputPacketSignal = static_cast< sc_core::sc_signal< counter_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< counter_t >) ));

        _packetSendCompletedSignal = static_cast< sc_core::sc_signal< counter_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< counter_t >) ));

        _outputPacketSizeSignal = static_cast< sc_core::sc_signal< size_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< size_t >) ));

        _outputPacketLabelSignal = static_cast< sc_core::sc_signal< label_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< label_t >) ));

        _newOutputPacketSignal = static_cast< sc_core::sc_signal< counter_t > * > (
			::operator new[] ( _PROXIES_NUMBER * sizeof(sc_core::sc_signal< counter_t >) ));


        // Calling ports constructors.
        for ( register unsigned int i = 0; i < _PROXIES_NUMBER; ++i )
        {
            std::string s;
            std::stringstream ss;
            ss << i;
            ss >> s;

            new (inputPacket + i) sc_core::sc_in< rtl_packet_t >(("inputPacket" + s).c_str());
            new (outputPacket + i) sc_core::sc_out< rtl_packet_t >(("outputPacket" + s).c_str());

            new (_inputPacketSignal + i) sc_core::sc_signal< rtl_packet_t >(("inputPacketSignal" + s).c_str());
            new (_outputPacketSignal + i) sc_core::sc_signal< rtl_packet_t >(("outputPacketSignal" + s).c_str());

            new (_carrierSignal + i) sc_core::sc_signal< carrier_t >(("carrierSignal" + s).c_str());
            new (_inputPacketSizeSignal + i) sc_core::sc_signal< size_t >(("inputPacketSizeSignal" + s).c_str());
            new (_inputPacketLabelSignal + i) sc_core::sc_signal< label_t >(("inputPacketLabelSignal" + s).c_str());
            new (_newInputPacketSignal + i) sc_core::sc_signal< counter_t >(("newInputPacketSignal" + s).c_str());
            new (_packetSendCompletedSignal + i) sc_core::sc_signal< counter_t >(("packetSendCompletedSignal" + s).c_str());
            new (_outputPacketSizeSignal + i) sc_core::sc_signal< size_t >(("outputPacketSizeSignal" + s).c_str());
            new (_outputPacketLabelSignal + i) sc_core::sc_signal< label_t >(("outputPacketLabelSignal" + s).c_str());
            new (_newOutputPacketSignal + i) sc_core::sc_signal< counter_t >(("newOutputPacketSignal" + s).c_str());
        }
    }


    template< const int PACKET_SIZE >
    void RtlTask_if_t< PACKET_SIZE >::delete_ports(
        sc_core::sc_port_base * p1, sc_core::sc_port_base * p2 )
    {
        sc_core::sc_in< rtl_packet_t > * in =
        static_cast<sc_core::sc_in< rtl_packet_t > *>(p1);

        sc_core::sc_out< rtl_packet_t > * out =
        static_cast<sc_core::sc_out< rtl_packet_t > *>(p2);

        delete in;
        delete out;
    }

  } }



#endif

